package VueControleur;

import java.awt.Color;
import java.awt.Font;
import java.awt.GridLayout;
import java.awt.Rectangle;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.geom.AffineTransform;
import java.awt.image.AffineTransformOp;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Observable;
import java.util.Observer;
import java.util.function.BooleanSupplier;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.imageio.ImageIO;
import javax.swing.BorderFactory;
import javax.swing.ImageIcon;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.SwingUtilities;
import javax.swing.border.Border;
import javax.swing.border.CompoundBorder;
import javax.swing.border.EmptyBorder;

import modele.Direction;
import modele.Fantome;
import modele.Jeu;
import modele.Menu;
import modele.Mur;
import modele.Entite;
import modele.ActionMenu;
import modele.Balle;
import modele.Pacman;
import modele.Object_Jeu;
import modele.Pac_Gomme;

import java.lang.Math;


/** Cette classe a deux fonctions :
 *  (1) Vue : proposer une représentation graphique de l'application (cases graphiques, etc.)
 *  (2) Controleur : écouter les évènements clavier et déclencher le traitement adapté sur le modèle (flèches direction Pacman, etc.))
 *
 * @author freder
 */
public class VueControleurPacMan extends JFrame implements Observer {

    private Jeu jeu; // référence sur une classe de modèle : permet d'accéder aux données du modèle pour le rafraichissement, permet de communiquer les actions clavier (ou souris)
    private Menu menu;
    
    private int sizeX; // taille de la grille affichée
    private int sizeY;

    private ImageIcon icoPacMan; // icones affichées dans la grille
    private ImageIcon icoFantome;
    private ImageIcon icoCouloir;
    private ImageIcon icoMur;
    private ImageIcon icoBalle;
    private ImageIcon icoVie;
    private ImageIcon icoPacGom;
    private ImageIcon icoSuperPacMan;

    private JLabel[][] tabJLabel; // cases graphique (au moment du rafraichissement, chaque case va être associé à une icône, suivant ce qui est présent dans la partie modèle)
    private JLabel[] lifesLabel;
    
    private JLabel textScore;
    
    public boolean quit;
    private boolean enJeu;
    
    BooleanSupplier mettreAJrAff;
    
    public VueControleurPacMan(int _sizeX, int _sizeY) {

        sizeX = _sizeX;
        sizeY = _sizeY;

        this.quit = false;
        this.enJeu = false;
        
        chargerLesIcones();


        setTitle("PacMan VS Coronavirus");
        setSize(sizeX*50, (sizeY+1) * 50);
        setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE); // permet de terminer l'application à la fermeture de la fenêtre

        ajouterEcouteurClavier();

    }

    private void ajouterEcouteurClavier() {

        addKeyListener(new KeyAdapter() { // new KeyAdapter() { ... } est une instance de classe anonyme, il s'agit d'un objet qui correspond au controleur dans MVC
            @Override
            public void keyPressed(KeyEvent e) {
                
            	if(enJeu)	
	                switch(e.getKeyCode()) {  // on écoute les flèches de direction du clavier
	                    case KeyEvent.VK_LEFT : jeu.getPacman().setDirection(Direction.gauche); break;
	                    case KeyEvent.VK_RIGHT : jeu.getPacman().setDirection(Direction.droite); break;
	                    case KeyEvent.VK_DOWN : jeu.getPacman().setDirection(Direction.bas); break;
	                    case KeyEvent.VK_UP : jeu.getPacman().setDirection(Direction.haut); break;
	                }
            	else
	                switch(e.getKeyCode()) {
	                    case KeyEvent.VK_DOWN : menu.action(ActionMenu.descendre); break;
	                    case KeyEvent.VK_UP : menu.action(ActionMenu.monter); break;
	                    case KeyEvent.VK_ENTER : menu.action(ActionMenu.entrer); break;
	                }
            		
                
            }

        });

    }

    public void setAffichage(Observable affichage) {
        if(affichage instanceof Jeu)
        {
        	jeu = (Jeu) affichage;
        	enJeu = true;
        	placerCompGraphJeu();
        	mettreAJrAff = this::mettreAJourAffichageJeu;
               
                
        }
        else if(affichage instanceof Menu)
        {
            menu = (Menu) affichage;
            this.enJeu = false;
            placerCompGraphMenu();
            mettreAJrAff = this::mettreAJourAffichageMenu;
        }
    }

    private void chargerLesIcones() {
        icoPacMan = chargerIcone("Images/Pacman.png");
        icoCouloir = chargerIcone("Images/Couloir.png");
        icoFantome = chargerIcone("Images/Fantome.png");
        icoMur = chargerIcone("Images/Mur.png");
        icoBalle = chargerIcone("Images/Balle.png");
        icoVie = chargerIcone("Images/Vie.png");
        icoPacGom = chargerIcone("Images/pac-gomme.png");
        icoSuperPacMan = chargerIcone("Images/Super-pacman.png");
    }

    private ImageIcon chargerIcone(String urlIcone) {

        BufferedImage image = null;
        try {
            image = ImageIO.read(new File(urlIcone));
        } catch (IOException ex) {
            Logger.getLogger(VueControleurPacMan.class.getName()).log(Level.SEVERE, null, ex);
            return null;
        }
        return new ImageIcon(image);
    }

    
    
    /// v Affichage Menu v
    
    private void placerCompGraphMenu() {
    	int lastSc = 0;
    	getContentPane().removeAll();
    	
        JComponent grilleJLabels = new JPanel(new GridLayout(2, 1)); // grilleJLabels va contenir les cases graphiques et les positionner sous la forme d'une grille
        JComponent grilleOptions = new JPanel(new GridLayout(4, 1));

        tabJLabel = new JLabel[1][menu.getNbOption()+1];

        JLabel jlab;
        
        if(jeu != null)
        {
        	jlab = new JLabel("Game Over !");
        	jlab.setForeground(Color.red);
        	jlab.setHorizontalAlignment(JLabel.CENTER);
        	jlab.setFont(new Font("Serif", Font.BOLD, 45));
        	grilleOptions.add(jlab);
        
        	lastSc = jeu.getScore();
        	jlab = new JLabel("Score : " + lastSc);
        	jlab.setHorizontalAlignment(JLabel.CENTER);
        	jlab.setFont(new Font("Serif", Font.PLAIN, 40));
        	grilleOptions.add(jlab);        	
        }
        else
        {      	
        	jlab = new JLabel("");
        	grilleOptions.add(jlab);
        	jlab = new JLabel("");
        	grilleOptions.add(jlab);
        }
        
        jlab = new JLabel("Jouer");
    	jlab.setHorizontalAlignment(JLabel.CENTER);
    	jlab.setFont(new Font("Serif", Font.PLAIN, 32));
        tabJLabel[0][0] = jlab;
        grilleOptions.add(jlab);

        jlab = new JLabel("Quitter");
    	jlab.setHorizontalAlignment(JLabel.CENTER);
    	jlab.setFont(new Font("Serif", Font.PLAIN, 32));
        tabJLabel[0][1] = jlab;
        grilleOptions.add(jlab);
        
        grilleJLabels.add(grilleOptions);

 
        HashMap<Integer,String> scores = menu.getScoreS();
        boolean needSave = false;
        JComponent grilScores = new JPanel(new GridLayout(10, 1));
        ArrayList<Integer> lstScore = new ArrayList<Integer>(scores.keySet());
        lstScore.sort(null);
        if(lstScore.get(0) <  lastSc && lstScore.contains(lastSc) == false)
        {
        	
        	scores.remove(lstScore.get(0));
        	String nameJ = (String)JOptionPane.showInputDialog(null,"Indiqué votre nom");
        	scores.put(lastSc, nameJ);
        	lstScore.add(lastSc);
        	lstScore.remove(0);
        	menu.saveScores(scores);
            lstScore.sort(null);
        }    
        
        for(int i = lstScore.size()-1; i>=0; i-- )
        {	
        	
            jlab = new JLabel(scores.get(lstScore.get(i))+". ");
        	jlab.setHorizontalAlignment(JLabel.RIGHT);
        	jlab.setFont(new Font("Comic Sans MS", Font.PLAIN, 20));
        	grilScores.add(jlab);
            jlab = new JLabel(String.valueOf(lstScore.get(i)));
        	jlab.setHorizontalAlignment(JLabel.LEFT);
        	jlab.setFont(new Font("Comic Sans MS", Font.PLAIN, 20));
        	grilScores.add(jlab);
        }
        grilleJLabels.add(grilScores);
        
        add(grilleJLabels);
    }
    
    /**
     * Il y a une grille du côté du modèle ( jeu.getGrille() ) et une grille du côté de la vue (tabJLabel)
     */
    private boolean mettreAJourAffichageMenu() {
    	
    	for(int i=0; i < menu.getNbOption(); i++)
    	{
    		tabJLabel[0][i].setBorder(BorderFactory.createEmptyBorder());
    	}
    	
    	tabJLabel[0][menu.option-1].setBorder( BorderFactory.createLineBorder(Color.BLUE, 10) );
    	Border border = tabJLabel[0][menu.option-1].getBorder();
    	Border margin = new EmptyBorder(10,10,10,10);
    	tabJLabel[0][menu.option-1].setBorder(new CompoundBorder(border, margin));
    	
    	if(menu.quit)
    		this.quit = true;
    	
        return true;

    }


    
    // v Affichage Menu v
    
    private void placerCompGraphJeu() {
    	
    	getContentPane().removeAll();
    	
        JComponent grilleJLabels = new JPanel(new GridLayout(sizeX+1, sizeY)); // grilleJLabels va contenir les cases graphiques et les positionner sous la forme d'une grille

        tabJLabel = new JLabel[sizeX][sizeY];
        

        for (int y = 0; y < sizeY; y++) {
            for (int x = 0; x < sizeX; x++) {

                JLabel jlab = new JLabel();
                tabJLabel[x][y] = jlab; // on conserve les cases graphiques dans tabJLabel pour avoir un accès pratique à celles-ci (voir mettreAJourAffichage() )
                grilleJLabels.add(jlab);
                
            }

        }

        grilleJLabels.add(new JLabel("Score"));
        grilleJLabels.add(new JLabel(":"));
        textScore = new JLabel("0");
        grilleJLabels.add(textScore);

        for (int x = 3; x < sizeY-3; x++)
            grilleJLabels.add(new JLabel(""));
        	

        lifesLabel = new JLabel[3];
        for(int i=2; i >= 0; i--)
        {
        	JLabel jlab = new JLabel();
        	lifesLabel[i]= jlab;
            grilleJLabels.add(jlab);
        }
      	
        add(grilleJLabels);
    }
    
    /**
     * Il y a une grille du côté du modèle ( jeu.getGrille() ) et une grille du côté de la vue (tabJLabel)
     */
    private boolean mettreAJourAffichageJeu() {
        
        for (int x = 0; x < sizeX; x++) {
            for (int y = 0; y < sizeY; y++) {
                tabJLabel[x][y].setIcon(null);
                Entite[][] grilEnt = jeu.getGestEnt().getGrille();
                if (grilEnt[x][y] instanceof Pacman) { // si la grille du modèle contient un Pacman, on associe l'icône Pacman du côté de la vue
                    
                    Pacman pm = (Pacman) grilEnt[x][y];
                    double rot = 0;
                    switch(pm.getDirection())
                    {
                    	case haut:
                    		rot = 3*Math.PI/2;
                    		break;
                    	case bas:
                    		rot = Math.PI/2;
                    		break;
                    	case gauche:
                    		rot = Math.PI;
                    		break;
                    	case droite:
                    	default:
                    		rot =2* Math.PI;
                    }
                    
                    
                    BufferedImage imgPlayer;

                    if(pm.invincible==true){
                    	imgPlayer = (BufferedImage) icoSuperPacMan.getImage();
                    }
                    else{
                    	imgPlayer = (BufferedImage) icoPacMan.getImage();
                    }
                    

                    AffineTransform tx = new AffineTransform();
                    tx.rotate(rot, imgPlayer.getWidth() / 2, imgPlayer.getHeight() / 2);

                    AffineTransformOp op = new AffineTransformOp(tx,
                        AffineTransformOp.TYPE_BILINEAR);
                    imgPlayer = op.filter(imgPlayer, null);
                        

                    tabJLabel[x][y].setIcon(new ImageIcon(imgPlayer));
       
                } else if (grilEnt[x][y] instanceof Fantome) {
                    
                    tabJLabel[x][y].setIcon(icoFantome);
                } 
                else if (jeu.getGrilleObject()[x][y] instanceof Mur) {
                    
                    tabJLabel[x][y].setIcon(icoMur);
                       
                }else if (jeu.getGrilleObject()[x][y] instanceof Balle) {
                    
                    tabJLabel[x][y].setIcon(icoBalle);
                       
                }
                else if (jeu.getGrilleObject()[x][y] instanceof Pac_Gomme) {
                    tabJLabel[x][y].setIcon(icoPacGom);
                       
                }            

            }
        }
        
        for(int i=0; i < 3; i++)
        {
        	if(jeu.getVie() >= i+1)
        		lifesLabel[i].setIcon(icoVie);
        	else
        		lifesLabel[i].setIcon(null);
        		
        }
        
        textScore.setText(String.valueOf(jeu.getScore()));
        
        return true;

    }

    @Override
    public void update(Observable o, Object arg) {
        
        
        this.mettreAJrAff.getAsBoolean();
        
    }

}

