package modele;

import java.awt.Point;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Scanner;

public class GestionEntite {
    
    private Pacman pm;
    private HashMap<Entite, Point> map = new  HashMap<Entite, Point>(); // permet de récupérer la position d'une entité à partir de sa référence
    
    private Entite[][] grilleEntites;
    private Fantome nwFantome;
    
    private Jeu jeu;
    
    
    public GestionEntite(Jeu _jeu)
    {
    	jeu = _jeu;
    }
    
    // Permet de changer les fantome qui apparaisse
	private Fantome newFantome(Point ptOrig)
	{
		if (nwFantome == null || nwFantome instanceof Clyde)
				return new Blinky(jeu, ptOrig);
		
		else if(nwFantome instanceof Blinky)
				return new Pinky(jeu, ptOrig);
		
		else if(nwFantome instanceof Pinky)
				return new Inky(jeu, ptOrig);
		
		else
			return new Clyde(jeu, ptOrig);
	}
	
    public void initialisationDesEntites() {
        grilleEntites = new Entite[jeu.SIZE_X][jeu.SIZE_Y];
        map = new  HashMap<Entite, Point>();
        nwFantome = null;
    }
    
    public void addFantome(int x, int y)
    {
        nwFantome = newFantome(new Point(x, y));
        grilleEntites[x][y] = nwFantome; 
        map.put(nwFantome, nwFantome.getPointOrigin());
    }
    
    public void addPacMan(int x, int y)
    {
        pm = new Pacman(jeu, new Point(x, y)); 
        grilleEntites[x][y] = pm;
        map.put(pm, new Point(x, y)); 
    }

    
    public void RestartEntites()
    {
    	Entite[][] grilleTmp = grilleEntites;
    	grilleEntites = new Entite[jeu.SIZE_X][jeu.SIZE_Y];
    	for(int x=0; x < jeu.SIZE_X; x++ )
        	for(int y=0; y < jeu.SIZE_Y; y++ )
        	{
        		if(grilleTmp[x][y] != null)
        		{
        			deplacerEntite(new Point(x, y),grilleTmp[x][y].getPointOrigin(),grilleTmp[x][y]);
        		}
        	}
    }

    
    public void respawnFan(Fantome fan)
    {
    	int x = map.get(fan).x;
        int y = map.get(fan).y;
		
        grilleEntites[x][y] = null;
        
        map.put(fan, fan.getPointOrigin());
        x = map.get(fan).x;
        y = map.get(fan).y;
        grilleEntites[x][y] = fan;
    }
    
    public Point calculerPointCible(Point pCourant, Direction d) {
        Point pCible = null;
        try
        {
            
            switch(d) {
                case haut: pCible = new Point(pCourant.x, pCourant.y - 1); break;
                case bas : pCible = new Point(pCourant.x, pCourant.y + 1); break;
                case gauche : if(pCourant.x - 1 >=0){
                    pCible = new Point(pCourant.x - 1, pCourant.y);
                }else pCible = new Point(20, pCourant.y);
                break;
                case droite :if(pCourant.x + 1<21){
                    pCible = new Point(pCourant.x + 1, pCourant.y);
                }else pCible = new Point(0, pCourant.y);
                break; 
            }
        }
        catch(NullPointerException e)
        {
        	return new Point(0,0);
        }
        
        return pCible;
    }
    
    public void deplacerEntite(Point pCourant, Point pCible, Entite e) {
        grilleEntites[pCourant.x][pCourant.y] = null;
        grilleEntites[pCible.x][pCible.y] = e;
        map.put(e, pCible);
    }
    
    /** Vérifie que p est contenu dans la grille
     */
    private boolean contenuDansGrille(Point p) {
        return p.x >= 0 && p.x < jeu.SIZE_X && p.y >= 0 && p.y < jeu.SIZE_Y;
    }
    
    public Entite entiteALaPosition(Point p) {
    	Entite retour = null;
        
        if (contenuDansGrille(p)) {
            retour = grilleEntites[p.x][p.y];
        }
        
        return retour;
    }
      

	public Entite[][] getGrille() {
		return grilleEntites;
	}
   
    public Pacman getPacman() {
        return pm;
    }

    public HashMap<Entite, Point> getMapEnt() {
        return map;
    }
    


    /// Fonction gestion IA fantome
    
    public Direction dirOppose(Direction d)
    {
        switch(d)
        {
        	case haut:
        		return Direction.bas;
        	case bas:
        		return Direction.haut;
        	case gauche:
        		return Direction.droite;
        	case droite:
        	default:
        		return Direction.gauche;
        		
        }
    	
    }
    
    // Savoir dans quelle direction peut se déplacer un fantome
    public Direction directionPossible(Entite e)
    {
        
        Point pCourant = map.get(e);
        Direction d = e.d;
    	Point pCible = calculerPointCible(pCourant, d);
    	// On va tout droit si c'est possible
    	if(!(entiteALaPosition(pCible) instanceof Fantome) &&
    			!(jeu.objectALaPosition(pCible) instanceof Mur)) 
    		return d;
        
        /* Sinon on voit ou on peut aller
         * Pour tester une direction x
         * 		=> On vérifie que c'est pas un demi tour
         * 		=> On vérifie si il y a pas de mur ou pas de fantome
         */
        
        
        // Test Haut
        if(e.d != Direction.bas)
        {
        	pCible = calculerPointCible(pCourant, Direction.haut);
        	if(!(entiteALaPosition(pCible) instanceof Fantome) &&
        			!(jeu.objectALaPosition(pCible) instanceof Mur)) 
        		return Direction.haut;
        }
        // Test Bas
        if(e.d != Direction.haut)
        {
        	pCible = calculerPointCible(pCourant, Direction.bas);
        	if(!(entiteALaPosition(pCible) instanceof Fantome) &&
        			!(jeu.objectALaPosition(pCible) instanceof Mur)) 
        		return Direction.bas;
        	
        }
        // Test gauche
        if(e.d != Direction.droite)
        {
        	pCible = calculerPointCible(pCourant, Direction.gauche);
        	if(!(entiteALaPosition(pCible) instanceof Fantome) &&
        			!(jeu.objectALaPosition(pCible) instanceof Mur)) 
        		return Direction.gauche;
        	
        }
        // Test droite
        if(e.d != Direction.gauche)
        {
        	pCible = calculerPointCible(pCourant, Direction.droite);
        	if(!(entiteALaPosition(pCible) instanceof Fantome) &&
        			!(jeu.objectALaPosition(pCible) instanceof Mur)) 
        		return Direction.droite;
        	
        }

        
        return dirOppose(d);
    }

    public ArrayList<Direction> dirDispo(Entite e)
    {
    	return dirDispo(map.get(e));
    }
    
    public ArrayList<Direction> dirDispo(Point pt)
    {
    	ArrayList<Direction> ld = new ArrayList<Direction>();

    	Point pCible;
    	
        // Test Haut
    	pCible = calculerPointCible(pt, Direction.haut);
    	if(!(jeu.objectALaPosition(pCible) instanceof Mur) && !(entiteALaPosition(pCible) instanceof Fantome)) 
    		ld.add(Direction.haut);

        // Test Bas
    	pCible = calculerPointCible(pt, Direction.bas);
    	if(!(jeu.objectALaPosition(pCible) instanceof Mur) && !(entiteALaPosition(pCible) instanceof Fantome))
    		ld.add(Direction.bas);
        // Test gauche
    	pCible = calculerPointCible(pt, Direction.gauche);
    	if(!(jeu.objectALaPosition(pCible) instanceof Mur) && !(entiteALaPosition(pCible) instanceof Fantome))
    		ld.add(Direction.gauche);
        // Test droite
    	pCible = calculerPointCible(pt, Direction.droite);
    	if(!(jeu.objectALaPosition(pCible) instanceof Mur) && !(entiteALaPosition(pCible) instanceof Fantome))
    		ld.add(Direction.droite);
    	
    	return ld;
    }
    
    public Direction goDerrieretPM(Entite e)
    {
    	try
    	{
	    	ArrayList<Direction> ld = dirDispo(e);

	    	if(ld.size() > 1)
	    		ld.remove(dirOppose(e.d));
	    	
	    	if(ld.size() == 1)
	    		return ld.get(0);
	    	
	    	Point ptF = map.get(e);
	    	Point ptPM = map.get(pm);
	    	if(ld.contains(Direction.bas) && ptF.y<ptPM.y)
	    		return Direction.bas;
	    	else if(ld.contains(Direction.haut) && ptF.y>ptPM.y)
	    		return Direction.haut;
	    	else if(ld.contains(Direction.droite) && ptF.x<ptPM.x)
	    		return Direction.droite;
	    	else if(ld.contains(Direction.gauche) && ptF.x>ptPM.x)
	    		return Direction.gauche;
	    		
	
			if(ld.contains(pm.d))
				return pm.d;

	    	if(ld.size() != 0)
	    		return ld.get(0);
			return directionPossible(e);
    	}
    	catch(NullPointerException exc)
    	{
			return directionPossible(e);
    	}
    		
    }

    public Direction goDevantPM(Entite e)
    {
    	try
    	{
	    	ArrayList<Direction> ld = dirDispo(e);
	    	if(ld.size() == 1)
	    		return ld.get(0);
	    
	    	ld.remove(dirOppose(e.d));
	    	
	    	if(ld.size() == 1)
	    		return ld.get(0);
	    
	    	
	    	Point ptF = map.get(e);
	    	Point ptPM = map.get(pm);
	    	
	    	switch(pm.d)
	    	{
	    		case bas:
	    			if(ptF.y>ptPM.y && ld.contains(Direction.haut))
	    	    		return Direction.haut;
	    		case haut:
	    			if(ptF.y<ptPM.y && ld.contains(Direction.bas))
	    	    		return Direction.bas;
	    		case gauche:
	    			if(ptF.x<ptPM.x && ld.contains(Direction.droite))
	    	    		return Direction.droite;
	    		case droite:
	    			if(ptF.x>ptPM.x && ld.contains(Direction.gauche))
	    	    		return Direction.gauche;
	    	
	    	}
	    	return goDerrieretPM(e);
    	}
    	catch(NullPointerException exc)
    	{
			return directionPossible(e);
    	}
    		
    }
    
    public void runEntite()
    {
    	for (Entite e : map.keySet()) { // déclenchement de l'activité des entités, map.keySet() correspond à la liste des entités
            e.run(); 
        }
        if(this.pm.invincible)
        {
        	this.pm.tempsInvincibilite -= 1;
        	if(this.pm.tempsInvincibilite == 0)
        		this.pm.invincible = false;
        }
    }
    
}
