/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package modele;

import java.awt.Point;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
 
import java.util.Observable;
import java.util.logging.Level;
import java.util.logging.Logger;


import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Scanner;
import java.io.File;
import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.SourceDataLine;
import javax.sound.sampled.UnsupportedAudioFileException;

/** La classe Jeu a deux fonctions 
 *  (1) Gérer les aspects du jeu : condition de défaite, victoire, nombre de vies
 *  (2) Gérer les coordonnées des entités du monde : déplacements, collisions, perception des entités, ... 
 *
 * @author freder
 */
public final class Jeu extends Observable implements Runnable {

    public static final int SIZE_X = 22;
    public static final int SIZE_Y = 22;
    
    private int vie = 3;
    private int score = 0;
    
    private boolean jeuEnding = false;
    
    private int nbBalle=0;
    private String level = "Ligne0";
    
    private Object_Jeu[][] grilleObject = new Object_Jeu[SIZE_X][SIZE_Y]; // permet de récupérer un object à partir de ses coordonnées
    
    private GestionEntite gestEnt;
    
    public Jeu() {
        gestEnt = new GestionEntite(this);
        this.restart();
    }
    
    private void restart()
    {
        this.score = 0;
        this.vie = 3;
        this.jeuEnding = false;
        this.newLevel();
    }
    
    private void newLevel()
    {
        try(Scanner s=new Scanner(new File("map/liste_niveau.txt"))){
          
            String line =s.nextLine();
            while(!(this.level.equals(line))){
                    line=s.nextLine();
            }if(s.hasNextLine()){
              line=s.nextLine();
              this.level=line;  
            }else{
                this.Level_Aleatoire();
            }


        }catch (IOException exception) {
            System.out.println("Je ne peux pas lire le fichier liste_niveau.txt");
        }  
        
        initialisationDesEntites();
        initialisationDesObjects();
        
    }
    
    private void Level_Aleatoire(){
        try(Scanner s=new Scanner(new File("map/liste_niveau.txt"))){
          
            while(this.level=="Ligne0"){
                for(int i=0; i<(int)(Math.random() * 4); i++){
                    this.level=s.nextLine();
                }
                
            }
            
        }catch (IOException exception) {
            System.out.println("Je ne peux pas lire le fichier liste_niveau.txt");
        }  
    }
    
    public Pacman getPacman() {
        return gestEnt.getPacman();
    }
    
     public Object_Jeu[][] getGrilleObject() {
        return grilleObject;
    }
    
    private void initialisationDesEntites() {
    	gestEnt.initialisationDesEntites();
        
        String fichier="map/"+this.level;
        try(Scanner scanner=new Scanner(new File(fichier))){
            while(scanner.hasNextLine()){
                
                Scanner line=new Scanner(scanner.nextLine());//Chaque ligne donne les coordonnées d'un bloc
                line.useDelimiter("/");
                int t=line.nextInt();//Recupère le premier int de la ligne
                int l = line.nextInt();//indice de la ligne de l'entite a ajouter
                int c = line.nextInt();//indice de la colonne de l'entite a ajouter
                if(t==2){ //Si la ligne commence par un "2" alors c'est un pacman
                    gestEnt.addPacMan(c, l);
                
                }else if(t==3){ //Si la ligne commence par un "3" alors c'est un fantome
                    gestEnt.addFantome(c, l);
                }
                line.close();
                
            }
            scanner.close();

        }catch (IOException exception) {
            System.out.println("Je ne peux pas lire le fichier fichier_map.txt pour l'initialisation des entites");
        }   
    }
 
    private void initialisationDesObjects() {
        grilleObject = new Object_Jeu[SIZE_X][SIZE_Y];
        
        Mur m = new Mur(this);
        Pac_Gomme pg = new Pac_Gomme(this);
        String fichier="map/"+this.level;
        try(Scanner scanner=new Scanner(new File(fichier))){
            while(scanner.hasNextLine())  //Lit le fichier ligne par ligne
            {
                Scanner line=new Scanner(scanner.nextLine());//Chaque ligne donne les coordonnées d'un bloc
                line.useDelimiter("/");
                int t=line.nextInt();//Recupère le premier int de la ligne
                if(t==1){ //Si la ligne commence par un "1" alors c'est un mur
                    
                    int l = line.nextInt();//indice de la ligne du mur ajouter
                    int c = line.nextInt();//indice de la colonne du mur ajouter
                
                    grilleObject[c][l] = m;//Créé le bloc
                    
                }else if(t==4){ //Si la ligne commence par un "4" alors c'est une pacgomme
                    
                    int l = line.nextInt();//indice de la ligne du mur ajouter
                    int c = line.nextInt();//indice de la colonne du mur ajouter
                
                    grilleObject[c][l] = pg;//Créé le bloc
                    
                }
                
                line.close();
            }
            scanner.close();
            
        }catch (IOException exception) {
            System.out.println("Je ne peux pas lire le fichier fichier_map.txt pour l'initialisation des objets");
        }
        
        //Complete le grille avec des balles
        Balle b = new Balle(this);
        
        for(int c=0; c<SIZE_X; c++){
            for(int l=0; l<SIZE_Y; l++){
                
                Point p = new Point(c, l);
                if(!(objectALaPosition(p) instanceof Mur) && gestEnt.entiteALaPosition(p) == null && !(objectALaPosition(p) instanceof Pac_Gomme)){
                    grilleObject[c][l] = b;
                    this.nbBalle+=1;
                }
            }
        }
    }

    private void musique() throws InterruptedException{
        AudioInputStream musique = null;
        SourceDataLine line;
                 
        try {
            musique = AudioSystem.getAudioInputStream(new File("musique/pacman.wav"));
        } catch (UnsupportedAudioFileException e) {
            e.printStackTrace();
            System.out.println("Fichier audio non lu");
        } catch (IOException e) {
            e.printStackTrace();
            System.out.println("Probleme à la lecture du fichier audio");
        }
         //Il est nécessaire de connaître le format audio du fichier d'entrée pour permettre à java de créer l'objet DataLine adéquat
         AudioFormat audioFormat = musique.getFormat();
         DataLine.Info info = new DataLine.Info(SourceDataLine.class,audioFormat);
          
        try {
            line = (SourceDataLine) AudioSystem.getLine(info);
                        
            } catch (LineUnavailableException e) {
              e.printStackTrace();
              return;
            }
          
        try {
                line.open(audioFormat);
        } catch (LineUnavailableException e) {
                    e.printStackTrace();
                    return;
        }
        line.start();
        
        try {
            byte bytes[] = new byte[1024];
            int bytesRead=0;
            while ((bytesRead = musique.read(bytes, 0, bytes.length)) != -1 && this.jeuEnding==false) {
                 line.write(bytes, 0, bytesRead);
                }
            if((musique.read(bytes, 0, bytes.length)) == -1){
                Thread.sleep(1000);
                this.musique();
            }
        } catch (IOException io) {
            io.printStackTrace();
            return;
        }  
        
        
    }
    
    private boolean contenuDansGrille(Point p) {
        return p.x >= 0 && p.x < SIZE_X && p.y >= 0 && p.y < SIZE_Y;
    }
    
     public Object_Jeu objectALaPosition(Point p) {
    	 Object_Jeu retour = null;
        
        if (contenuDansGrille(p)) {
            retour = grilleObject[p.x][p.y];
        }
        
        return retour;
    }
     

     public boolean deplacerEntite(Entite e, Direction d) {
         
         Point pCourant = gestEnt.getMapEnt().get(e);
         
         Point pCible = gestEnt.calculerPointCible(pCourant, d);
         
         // Si Elle rencontre une autre entité
         if(gestEnt.entiteALaPosition(pCible) != null)
         {
         	Entite entite = gestEnt.entiteALaPosition(pCible);
         	// Si c'est deux fantome Ils bougent pas
         	// Leur fonction de changement de direction les fais changer de direction
         	if(e instanceof Fantome && entite instanceof Fantome)
         	{
             	return false;
         	}
         	// Si Pacman rencontre un fantome c'est le drame
         	if(e instanceof Fantome && entite instanceof Pacman
         			|| entite instanceof Fantome && e instanceof Pacman)
         	{
         		Pacman pcm;
         		Fantome fan;
         		if(e instanceof Pacman)
         		{
         			pcm = (Pacman) e;
         			fan = (Fantome) entite;
         		}
         		else
         		{
         			pcm = (Pacman) entite;
     				fan = (Fantome) e;
         		}
             	// Ou pas
             	if(pcm.invincible)
             	{
         			this.score += 40;
         			
         			// Réinitialisation du fantome
                     gestEnt.respawnFan(fan);
                     
                     if(e instanceof Pacman)
                     {
                    	 gestEnt.deplacerEntite(pCourant, pCible, e);	
 	                    return true;
                     }
                     return false;
             	}
             		
         		this.vie -= 1;
         		if(this.vie == 0)
         			this.jeuEnding = true;
         		gestEnt.RestartEntites();
             	return false;
         	}

         	gestEnt.deplacerEntite(pCourant, pCible, e);
            return true;
         }
         // Le chemin est libre
         else if (contenuDansGrille(pCible) && !(objectALaPosition(pCible) instanceof Mur))
         {

             // Si il y a un object mangeable et que c'est pac-man
             if (e  instanceof Pacman &&
             		objectALaPosition(pCible) != null )
             {
             	Pacman pcm = (Pacman) e;
             	
             	Object_Jeu obj = objectALaPosition(pCible);
             	if(obj.mangeable)
             	{
             		if(obj instanceof Pac_Gomme)
             		{
             			this.score += 20;
             			pcm.invincible = true;
             			pcm.tempsInvincibilite = 15;
             			
             		}
             		else if(obj instanceof Balle)
             		{
             			this.score += 10;
                                 this.nbBalle -= 1;
             		}
             		grilleObject[pCible.x][pCible.y] = null;
             	}
             	else
             		return false;
             }
             
             
             gestEnt.deplacerEntite(pCourant, pCible, e);
             return true;
             
         }
         else 
             return false;
     }
  
    
    /**
     * Un processus est créé et lancé, celui-ci execute la fonction run()
     */
    public void start() throws InterruptedException {
        
    	if(this.jeuEnding)
    		this.restart();
    	
       new Thread(this).start();
       this.musique();
    }

    @Override
    public void run() {
        while (!this.jeuEnding) {

            gestEnt.runEntite();
            if(this.nbBalle==0){
                newLevel();
            }

            setChanged();
            notifyObservers(); // notification de l'observer pour le raffraichisssement graphique

            try {
                Thread.sleep(400); // pause de 0.4s
                
            } catch (InterruptedException ex) {
                Logger.getLogger(Pacman.class.getName()).log(Level.SEVERE, null, ex);
            }

        }
    }
    
    public int getVie() {
    	return vie;
    }

    public int getScore() {
    	return score;
    }
    
    
    public boolean getEnding()
    {
    	return this.jeuEnding;
    }
    
    public GestionEntite getGestEnt()
    {
    	return gestEnt; 
    }
}
